package edu.georgetown.gucs.dictionary;

import java.io.File;
import java.util.Collections;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Map.Entry;

import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.IgnoreCaseComparator;
import edu.georgetown.gucs.utility.Token;
import edu.georgetown.gucs.utility.FileLister;

/**
 * Thread for processing a file to add to a {@linkplain Dictionary}
 * 
 * @author Clay Shields
 * 
 **/
class DictionaryWorker extends Thread {

    private Dictionary parent;
    private List<String> tokenizerNames;
    private TokenizerList tokenizers;
    private FileLister fileLister;
    private File poison;

    /**
     * Constructor that specifies a file to tokenize and add to an existing dictionary
     * 
     * @param file
     *          the file to tokenize and add to this dictionary
     * @param tokenizers
     *          the list containing the names of the tokenizers to use on this file
     * @param dictionary
     *          the dictionary to add the file to
     */
    DictionaryWorker(List<String> tokenizerNames, Dictionary parent, FileLister fileLister, File poison) {
	this.tokenizerNames = tokenizerNames;
	this.parent = parent;
	this.fileLister = fileLister;
	this.poison = poison;
    }

    /**
     * Adds a file to an existing dictionary using the given tokenizers
     */
    @Override
	public void run() {
	Global.monitor.writeToBuffer("DictionaryWorker", "run_start", "Starting");

	boolean done = false;

	// Create a tokenizer list from the names of tokens provided
	this.tokenizers = new TokenizerList(tokenizerNames);
	

	while (!done){	    
	    
	    // process files that come from the fileLister 
	    File file = fileLister.takeFile();

	    if (file == poison){
		done = true;
		fileLister.putPoison(poison);
		Global.monitor.writeToBuffer("DictionaryWorker", "run_start", "Poison received, shutting down");
		return;
	    }
	    
	    Global.monitor.writeToBuffer("DictionaryWorker", "run_start", "Thread " + Thread.currentThread().getName() + "processing " + file);

	    // create a map to hold all the items
	    Map<Token, DictionaryEntry> newDictionary = Collections.synchronizedMap(new HashMap<Token, DictionaryEntry>());
	    

	    for (Entry<String, List<Token>> splitterEntry : this.tokenizers.tokenizeFile(file.getAbsolutePath()).entrySet()) {
		for (Token token : splitterEntry.getValue()) {
		    if (token.name().length() > 0) {
			if (newDictionary.containsKey(token)) {
			    newDictionary.get(token).incrementFrequencyCount();
			} else {
			    DictionaryEntry dictEntry = new DictionaryEntry(token);
			    newDictionary.put(token, dictEntry);
			}
		    }
		}
		
		if (newDictionary.size() > 0) {
		    
		    this.parent.mergeDictionary(newDictionary);
	
	    } else {
		Global.monitor.printErrorMessage("DictionaryWorker", "No tokens added to dictionary from " + file);
		Global.monitor.writeToBuffer("DictionaryWorker", "run", "No tokens added to dictionary from " + file);
	    }
        }
	}
    }
}