package edu.georgetown.gucs.dictionary;

import java.io.File;
import java.util.Collections;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Map.Entry;
import java.util.concurrent.LinkedBlockingQueue;


import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.IgnoreCaseComparator;
import edu.georgetown.gucs.utility.Token;
import edu.georgetown.gucs.utility.FileLister;

/**
 * Thread for processing a file to add to a {@linkplain Dictionary}
 * 
 * @author Clay Shields
 * 
 **/
class PartialDictionaryWorker extends Thread {

    private Dictionary parent;
    private List<String> tokenizerNames;
    private TokenizerList tokenizers;
    private LinkedBlockingQueue<File> sharedQueue = new LinkedBlockingQueue<File>();
    private File poison;


    /**
     * Constructor that specifies a file to tokenize and add to an existing dictionary
     * 
     * @param file
     *          the file to tokenize and add to this dictionary
     * @param tokenizers
     *          the list containing the names of the tokenizers to use on this file
     * @param dictionary
     *          the dictionary to add the file to
     */
    PartialDictionaryWorker(List<String> tokenizerNames, Dictionary parent, LinkedBlockingQueue<File> sharedQueue,
			    File poison) {
	this.tokenizerNames = tokenizerNames;
	this.sharedQueue = sharedQueue;
	this.parent = parent;	
	this.poison = poison;
    }

    /**
     * Adds a file to an existing dictionary using the given tokenizers
     */
    @Override
	public void run() {
	Global.monitor.writeToBuffer("PartialDictionaryWorker", "run_start", "Starting");

	boolean done = false;
	File file = null;

	// Create a tokenizer list from the names of tokens provided
	this.tokenizers = new TokenizerList(tokenizerNames);
	       
	while (!done){	    
	    
	    // process files that come from the parent
	    try {
		file = sharedQueue.take();
	    }
	    catch (InterruptedException e){
		ExceptionHandler.handleExceptionExit(e);
	    }

	    if (file == poison){
		done = true;
		try {
		    sharedQueue.put(poison);
		}
		catch (InterruptedException e){
		    ExceptionHandler.handleExceptionExit(e);
		}
		
		Global.monitor.writeToBuffer("PartialDictionaryWorker", "run_start", "Poison received, shutting down");
		return;
	    }

	    Global.monitor.writeToBuffer("PartialDictionaryWorker", "run_start", "Thread " + Thread.currentThread().getName() + "processing " + file);

	    // create a map to hold all the items
	    Map<Token, DictionaryEntry> newDictionary = new HashMap<Token, DictionaryEntry>();
	    

	    for (Entry<String, List<Token>> splitterEntry : this.tokenizers.tokenizeFile(file.getAbsolutePath()).entrySet()) {
		for (Token token : splitterEntry.getValue()) {
		    if (token.name().length() > 0) {
			if (newDictionary.containsKey(token)) {
			    newDictionary.get(token).incrementFrequencyCount();
			} else {
			    DictionaryEntry dictEntry = new DictionaryEntry(token);
			    newDictionary.put(token, dictEntry);
			}
		    }
		}
		
		if (newDictionary.size() > 0) {
		    
		    this.parent.mergeDictionary(newDictionary);
	
	    } else {
		Global.monitor.printErrorMessage("PartialDictionaryWorker", "No tokens added to dictionary from " + file);
		Global.monitor.writeToBuffer("PartialDictionaryWorker", "run", "No tokens added to dictionary from " + file);
	    }
        }
	}
    
    }
}