package edu.georgetown.gucs.experiment;

import java.io.File;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.List;
import java.util.Set;
import java.util.Random;

import java.util.concurrent.LinkedBlockingQueue;
import edu.georgetown.gucs.matcher.ExactFingerprintMatcher;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Pair;
import edu.georgetown.gucs.utility.Token;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.utility.ObjectBuilder;
import edu.georgetown.gucs.matcher.FingerprintMatcher;

/**
 * Uses a given set of parameters to create a dictionary and fingerprints of a set of documents. Matches are based on
 * comparisons of unmangled and mangled versions of each file and optionally stores the matches in a database.
 * 
 * @author Lindsay Neubauer
 * @author Clay Shields
 * 
 */
class Trial extends Thread {

    private int trialID = -1;
    private TrialParameters base_parameters;
    private boolean finished = false;
    private TrialParameters poison;
    private Experiment parent;
    private String threadName;
    private int dictionarySize = 0;
    

    private int experimentID;
    private boolean insert;
    private DBInterface database;
    private String dataset;
    private FileLister fileSource;
    private List<String> tokenizerNames;
    private double dictionarySampleSize;
    private double sampleSize;
    private Pair<Double,Double> idfRange;
    private String fingerprinterName;
    private String matcherName;
    private int minScore;
    private String mangler;
    private String splitter;
    private Random rnd;
    private Fingerprinter fingerprinter;
    


    private Set<TrialParameters> unmangledFingerprints = new HashSet<TrialParameters>();
    private Set<TrialParameters> mangledFingerprints = new HashSet<TrialParameters>();
    private Map<String, Double[]> manglerResults;
    private LinkedBlockingQueue<TrialParameters> parameterQueue;

    

    /**
     * Blank default constructor
     * 
     */
    
    Trial (){
	this.base_parameters = null;
	this.experimentID = 0;
	this.insert = false;
	this.database = null;
	this.dataset = null;
	this.fileSource = null;
	this.tokenizerNames = null;
	this.dictionarySampleSize = 0;
	this.sampleSize = 0;
	this.idfRange = null;
	this.fingerprinterName = null;
	this.matcherName = null;
	this.minScore = 0;
	this.mangler = null;
	this.splitter = null;
	this.rnd = null;
	this.fingerprinter = null;
	
    }
    


    /**
     * Constructor for threading
     * 
     * @param theQueue
     *    the queue from which the thread will draw parameters to process
     * @param thePosion
     *    an empty TrialParameters object that marks the end of the queue
     *
     */
    
    Trial (LinkedBlockingQueue<TrialParameters> theQueue, TrialParameters thePoison, Experiment parent){
	this.parameterQueue = theQueue;
	this.poison = thePoison;	
	this.parent = parent;
    }

    /**
     * Constructor that sets default values for this trial
     * 
     * @param trialParameters
     *          the trialParameters containing the default values for this trial
     */
    Trial(TrialParameters trialParameters) {
	this.base_parameters = trialParameters;
	this.experimentID = base_parameters.getExperimentID();
	this.insert = base_parameters.isInsert();
	this.database = base_parameters.getDatabase();
	this.dataset = base_parameters.getDataset();
	this.fileSource = base_parameters.getFileSource();
	this.tokenizerNames = base_parameters.getTokenizerNames();
	this.dictionarySampleSize = base_parameters.getDictionarySampleParameter();
	this.sampleSize = base_parameters.getSampleParameter();
	this.idfRange = base_parameters.getIDFRange();
	this.fingerprinterName = base_parameters.getFingerprinterName();
	this.matcherName = base_parameters.getMatcherName();
	this.minScore = base_parameters.getMinScore();
	this.mangler = base_parameters.getMangler();
	this.splitter = base_parameters.getSplitter();
	this.rnd = base_parameters.getRandom();
	this.fingerprinter = null;
    }


    /**
     * Sets values for this trial
     * 
     * @param trialParameters
     *          the trialParameters containing the default values for this trial
     */
    
    public void setTrialParameters (TrialParameters trialParameters){
	this.base_parameters = trialParameters;
    }

    /**
     * Provides the parameters for this trial that are needed for results comparisons
     * 
     * @return the trialParameters containing the relevant parameters
     */
    public TrialParameters getTrialParameters() {
	return this.base_parameters;
    }

    /**
     * Provides the results for each mangler; sleeps until the trial is finished
     * 
     * @return the map containing the manglers results from this trial
     */
    public Map<String, Double[]> getManglerResults() {
	while (!this.finished) {
	    try {
		Thread.sleep(1000);
	    } catch (InterruptedException e) {
		ExceptionHandler.handleExceptionExit(e);
	    }
	}
	return this.manglerResults;
    }

    /**
     * Runs the trial
     */
    @Override
    public void run() {
      
	threadName = Thread.currentThread().getName();
		

	boolean done = false;
      
	while (!done) {

	    // Make some clean space for the results
	    manglerResults = new LinkedHashMap<String, Double[]>();
	    unmangledFingerprints = new HashSet<TrialParameters>();
	    mangledFingerprints = new HashSet<TrialParameters>();
    

	    // Grab a set of parameters from the queue
	    try {
		base_parameters = parameterQueue.take();
	    } catch (InterruptedException e) {
		ExceptionHandler.handleExceptionExit(e);
	    }


	    if (this.base_parameters == null){
		Global.monitor.printMessage("Trial " + threadName,"Thread called with null parameters");
		throw new RuntimeException("Trial Thread called with null parameters");
	    }
	    
	    // Unpack the parameters
	    this.experimentID = base_parameters.getExperimentID();
	    this.insert = base_parameters.isInsert();
	    this.database = base_parameters.getDatabase();
	    this.dataset = base_parameters.getDataset();
	    this.fileSource = base_parameters.getFileSource();
	    this.tokenizerNames = base_parameters.getTokenizerNames();
	    this.dictionarySampleSize = base_parameters.getDictionarySampleParameter();
	    this.sampleSize = base_parameters.getSampleParameter();
	    this.idfRange = base_parameters.getIDFRange();
	    this.fingerprinterName = base_parameters.getFingerprinterName();
	    this.matcherName = base_parameters.getMatcherName();
	    this.minScore = base_parameters.getMinScore();
	    this.mangler = base_parameters.getMangler();
	    this.splitter = base_parameters.getSplitter();
	    this.rnd = base_parameters.getRandom();
	    this.fingerprinter = null;
	    



	    // Quit if we get the poison, and pass it on and exit
	    if (base_parameters == poison){
		done = true;
		try {
		    parameterQueue.put(poison);
		} catch (InterruptedException e) {
		    ExceptionHandler.handleExceptionExit(e);
		}
		done = true;
		return;
	    }	  

	    if (insert) {
		// The next call for newTrial used to use a random seed, just set to zero until we strip
		// out all the stuff about random seeds
		this.trialID = database.newTrial(0);
		database.addTrialToExperiment(this.base_parameters.getExperimentID(), this.trialID);
	    }

	    Global.monitor.printMessage("Trial", "Getting sublists from fileLister: " + fileSource.getDirectory());
					    
	    Pair<List<File>,List<File>> dictAndSampleLists = fileSource.getSublists(dictionarySampleSize,sampleSize);

	    List<File> dictionaryList = dictAndSampleLists.getA();
	    List<File> sampleList = dictAndSampleLists.getB();

	    Global.monitor.printMessage("Trial " + threadName, "Setting up dictionary");
	    // FIX - find a better way to work out the number of threads to use
	    int dictThreads = 4;
	    Dictionary dictionary = new Dictionary();
	    dictionary.threadingOn(dictThreads);

	    if (insert) {
		int dictionaryID = database.newDictionary(dictionary);
		database.addDictionaryToTrial(this.trialID, dictionaryID);
	    }	   

	    dictionary.setTokenizers(tokenizerNames);
	    dictionary.createPartialDictionary(dictionaryList, fileSource.getDirectory());

	    if ((idfRange.getA() > 0) || (idfRange.getB() < 1)) {
		dictionary.trimByIDF(idfRange.getA(), idfRange.getB());
	    }
	    this.dictionarySize = dictionary.size();
					    
	    Global.monitor.printMessage("Trial " + threadName, "Setting up fingerprints");	    

	    fingerprinter = (Fingerprinter) ObjectBuilder.buildObject(fingerprinterName);
	    fingerprinter.setDictionary(dictionary);

	    Global.monitor.printMessage("Trial " + threadName, "Computing null fingerprints for "
					+ sampleList.size() + " files");
	    Global.monitor.printMessage("Trial " + threadName, "Mangler is: " + fingerprinter.getMangler());
	    computeFingerprints(null, null, sampleList);

	    // Clay - made this only compute additional fingerprints iff 
	    // the mangler isn't null, since we already computed the null ones above
	    // Lost the algorithm, so I don't know if this is right.
	    if ((mangler != null) && (splitter != null)) {
		    fingerprinter.setManglerRNG(rnd);
		    fingerprinter.setMangler(mangler,dictionary);
		    fingerprinter.setSplitter(splitter);
		    		    
		    Global.monitor.printMessage("Trial " + threadName, 
						"Computing splitter and mangler fingerprints:  splitter: "
						+ splitter + "  mangler(s)=" + mangler);
		    Global.monitor.printMessage("Trial " + threadName, "Mangler is: " + fingerprinter.getMangler());
		    computeFingerprints(mangler, splitter, sampleList);
	    }


	    Global.monitor.printMessage("Trial " + threadName, "doComparisons");
	    doComparisons();
	    Global.monitor.printMessage("Trial " + threadName, "calculateTrialResults");
	    calculateTrialResults();
	    parent.addTrialResults(base_parameters, this.manglerResults);
	}
	return;
    }

    /**
     * Computes fingerprints for each of the given files
     * 
     * @param manglerSettings
     *          the string containing the mangler settings to use; null if taking unmangled fingerprints
     * @param splitterString
     *          the string containing the splitter to use; null if not splitting files
     */
    private void computeFingerprints(String manglerSettings, String splitterString, List<File> sampleFiles) {
	// rethinking this
	// aglorithm should be:
	// take a unmangled fingerprint with the splitter
	// (Note that fingeprints contain mulitple parts if the splitter is there)
	// take a mangled fingerprint with the splitter
	// Compare the mangled and unmangled fingerprints of the same tokens
	// 


	boolean unmangled = false;
	if (manglerSettings == null) {
	    unmangled = true;
	    manglerSettings = "null";
	}

	Global.monitor.printMessage("Trial " + threadName, "Computing fingerprints for mangler " + manglerSettings);
	Global.monitor.printMessage("Trial " + threadName, "Computing fingerprints for splitter " + splitterString);
	int fileCounter = 0;
	for (File file : sampleFiles) {
	    fileCounter++;
	    for (Pair<byte[], Token> fingerprintPair :fingerprinter.computeFingerprint(file.getAbsolutePath())) {
		if (this.insert) {
		    int currentManglerID = this.database.getManglerID(manglerSettings);
		    this.database.addManglerToTrial(this.trialID, currentManglerID);
		    this.base_parameters.getDatabase().insertFingerprint(file, fingerprintPair.getA(), this.trialID,
									 currentManglerID, splitterString);
		}
        
		TrialParameters trialObject = new TrialParameters(this.base_parameters);
		// Trial object has a file token and fingerprint
		trialObject.set(fingerprintPair.getB(), fingerprintPair.getA());

		if (unmangled) {
		    this.unmangledFingerprints.add(trialObject);
		} else {
		    this.mangledFingerprints.add(trialObject);
		}
	    }
	}
	Global.monitor.writeToBuffer("Trial " + threadName, "computeFingerprints", fileCounter + " files");
	Global.monitor.printMessage("Trial " + threadName, "computeFingerprints: " + fileCounter + " files");
    }

    /**
     * Compares each mangled and unmangled fingerprint using this trials matcher and an exactFingerprintMatcher and
     * records true-positive, false-positive, true-negative, and false-negative results based on this comparison
     */
    private void doComparisons() {
	Global.monitor.writeToBuffer("Trial " + threadName, "doComparisons", "mangled fingerprints: " + this.mangledFingerprints.size()
				     + "  unmangled fingerprint: " + this.unmangledFingerprints.size());
	Global.monitor.printMessage("Trial " + threadName, "Comparing " + this.mangledFingerprints.size() + " mangled fingerprints to "
				    + this.unmangledFingerprints.size() + " unmangled fingerprints");

	FingerprintMatcher matcher = (FingerprintMatcher) ObjectBuilder.buildObject(this.matcherName);
	// Clay - matcher parameter is never set here!?

	ExactFingerprintMatcher exactMatcher = new ExactFingerprintMatcher();
	for (TrialParameters mangledParameters : this.mangledFingerprints) {
	    byte[] mangledFP = mangledParameters.getFingerprint();
	    String mangledFile = mangledParameters.getFileToken().name();

	    // Not clear what state the fingerprinter is in here. Nor what it should be.
	    for (Pair<byte[], Token> withoutMangle :this.fingerprinter.computeFingerprint(mangledFile)) {
		for (TrialParameters unmangledParameters : this.unmangledFingerprints) {
		    byte[] unmangledFP = unmangledParameters.getFingerprint();

		    // Match and compare to an ExactFingerprint match of the unmangled fingerprints from each file
		    boolean matchClaim = matcher.match(unmangledFP, mangledFP);
		    boolean exactMatch = exactMatcher.match(unmangledFP, withoutMangle.getA());

		    // Set up results
		    Double[] result = this.manglerResults.get(mangledParameters.getMangler());
		    if (result == null) {
			result = new Double[] { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
		    }

		    // Update results
		    if (matchClaim == true && exactMatch == true) {
			result[Global.TRUE_POSITIVE]++;
		    } else if (matchClaim == true && exactMatch == false) {
			result[Global.FALSE_POSITIVE]++;
		    } else if (matchClaim == false && exactMatch == false) {
			result[Global.TRUE_NEGATIVE]++;
		    } else if (matchClaim == false && exactMatch == true) {
			result[Global.FALSE_NEGATIVE]++;
		    }
		    result[Global.DICTIONARY_SIZE] = (double) this.dictionarySize;
		    this.manglerResults.put(mangledParameters.getMangler(), result);

		    if (this.base_parameters.isInsert()) {
			int unmangledID = Integer.valueOf(unmangledParameters.toString().substring(
												   unmangledParameters.toString().charAt('@') + 1));
			int mangledID = Integer.valueOf(mangledParameters.toString().substring(
											       mangledParameters.toString().charAt('@') + 1));
			int score = matcher.getScore(unmangledFP, mangledFP);
			this.database.addToFingerprintMatch(this.trialID, unmangledID, mangledID, matchClaim,
										 exactMatch, score);
		    }
		}
	    }
	}
    }

    /**
     * Computes the precision, recall, and f-score for each mangler
     */
    private void calculateTrialResults() {
	Global.monitor.writeToBuffer("Trial " + threadName, "calculateTrialResults", "");
	for (TrialParameters mangledParameters : this.mangledFingerprints) {
	    Double[] result = this.manglerResults.get(mangledParameters.getMangler());

	    // precision == ppv
	    if ((result[Global.TRUE_POSITIVE] + result[Global.FALSE_POSITIVE]) > 0.0) {
		result[Global.PRECISION] = result[Global.TRUE_POSITIVE]
		    / (result[Global.TRUE_POSITIVE] + result[Global.FALSE_POSITIVE]);
	    } else {
		result[Global.PRECISION] = 0.0;
	    }

	    // recall == sensitivity
	    if ((result[Global.TRUE_POSITIVE] + result[Global.FALSE_NEGATIVE]) > 0.0) {
		result[Global.RECALL] = result[Global.TRUE_POSITIVE]
		    / (result[Global.TRUE_POSITIVE] + result[Global.FALSE_NEGATIVE]);
	    } else {
		result[Global.RECALL] = 0.0;
	    }

	    // harmonic mean of precision and recall
	    if ((result[Global.PRECISION] + result[Global.RECALL]) > 0) {
		result[Global.FSCORE] = 2 * ((result[Global.PRECISION] * result[Global.RECALL]) / (result[Global.PRECISION] + result[Global.RECALL]));
	    } else {
		result[Global.FSCORE] = 0.0;
	    }

	    if (this.base_parameters.isInsert()) {
		Global.monitor.writeToBuffer("Trial " + threadName, "calculateTrialResults", "insert into database");
		int mangledID = Integer.valueOf(mangledParameters.toString().substring(
										       mangledParameters.toString().charAt('@') + 1));

		this.base_parameters.getDatabase().addTrialResults(this.trialID, mangledID,
								   result[Global.TRUE_POSITIVE].intValue(), result[Global.TRUE_NEGATIVE].intValue(),
								   result[Global.FALSE_POSITIVE].intValue(), result[Global.FALSE_NEGATIVE].intValue(),
								   result[Global.PRECISION], result[Global.RECALL], result[Global.FSCORE]);
	    }
	    this.manglerResults.put(mangledParameters.getMangler(), result);
	}

	this.finished = true;
	if (this.base_parameters.isInsert()) {
	    Global.monitor.printMessage("Trial " + threadName, "Trial " + this.trialID + " Finished");
	} else {
	    Global.monitor.printMessage("Trial " + threadName, "Trial Finished");
	}
    }
}