package edu.georgetown.gucs.fingerprinter;

import java.io.File;
import java.util.Map;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;
import java.util.Collections;
import java.util.ArrayList;
import java.util.Random;
import java.math.BigInteger;

import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.ObjectBuilder;
import edu.georgetown.gucs.utility.Pair;
import edu.georgetown.gucs.utility.Token;


/**
 * This is a class that allows threaded parallel creation of fingerprints.
 * 
 * @author Clay Shields
 **/


public class BulkFingerprinter {

    private String fingerprintType;
    private FileLister fileSource;
    private Dictionary dictionary;
    private int maxThread;
    private List<Pair<byte[], Token>> fpList;
    private int finishedThreadCount;

    /**
     * A constructor for when you want to limit the number of threads that are run
     * @param fingerprinterName
     *           the name of the fingerprinter-derived class to use for fingerprinting
     * @param files
     *       A FileLister object instantiated with the directory containing the files to fingerprint
     * @param dictionary
     *      The dictionary to be used for fingerprinting
     * @param maxThread
     *      limit on number of threads to be used
     */

    public BulkFingerprinter(String fingerprinterName, FileLister files, Dictionary dictionary, int maxThread){

	this.fingerprintType = fingerprinterName;
	this.fileSource = files;
	this.dictionary = dictionary;
	this.maxThread = maxThread;
	this.finishedThreadCount = 0;
	fpList = new ArrayList<Pair<byte[], Token>>();
    }


    /**
     * A constructor that will set the number of threads equal to the number of cores.
     * @param fingerprinterName
     *           the name of the fingerprinter-derived class to use for fingerprinting
     * @param files
     *       A FileLister object instantiated with the directory containing the files to fingerprint
     * @param dictionary
     *      The dictionary to be used for fingerprinting
     */


    public BulkFingerprinter(String fingerprinterName, FileLister files, Dictionary dictionary){

	this.fingerprintType = fingerprinterName;
	this.fileSource = files;
	this.dictionary = dictionary;
	this.maxThread = Runtime.getRuntime().availableProcessors();
	this.finishedThreadCount = 0;
	fpList = new ArrayList<Pair<byte[], Token>>();
    }


    /**
     * Allow a thread to return its results. This is used instead of a synchronized
     * data structure for performance reasons - we want to have many threads read at
     * once for matching
     *
     */

    synchronized void putResults(List<Pair<byte[], Token>> incoming){

	Iterator<Pair<byte[], Token>> fpIterator = incoming.iterator();
	
	while(fpIterator.hasNext()){
	    fpList.add(fpIterator.next());
	}
	
	this.finishedThreadCount++;	
    }

  /**
   * Recursively computes all the fingerprints of documents in a given directory as 
   * Base64 encoded strings
   * 
   * @param base
   *          the directory to fingerprint
   * 
   * @return the Base64 encoded string containing the fingerprint and if applicable the start and end bytes for a split
   *         file
   */
    public List<Pair<byte[], Token>> computeAllFingerprints(String fingerprinterName, Dictionary dictionary, FileLister base) {
      
	//System.err.println("Entering computeAllFingerprints with max threads: " + maxThread);	     

	Global.monitor.writeToBuffer("BulkFingerprinter", "computeAllFingerprints", "Starting with " + maxThread + " threads");

      	ArrayList<BulkFingerprinterWorker> threads = new ArrayList<BulkFingerprinterWorker>(maxThread);
	File poison = new File(new BigInteger(128, new Random()).toString());
	
	//System.err.println("BulkFingeprinter: starting threads");

	for (int i = 0; i < maxThread; i++){
	    threads.add(i,new BulkFingerprinterWorker(fingerprinterName, dictionary, this, base, poison));
	    threads.get(i).setName("Bulk Fingerprint Worker " + i);
	    threads.get(i).setDaemon(true);
	    threads.get(i).start();
	}    

	//System.err.println("BulkFingeprinter: waiting");

	// Wait for the file lister to finish loading files
	while (!base.getLoaded()){
	    try {
		Thread.sleep(100);
	    }
	    catch (InterruptedException e){
		ExceptionHandler.handleExceptionExit(e);
	    }
	}

	// Now wait for the dictionary worker threads to finish
	boolean done = false;
	while (!done){
	    // Sleep a bit while the threads run
	    try {
		Thread.sleep(100);
	    }
	    catch (InterruptedException e){
		ExceptionHandler.handleExceptionExit(e);
	    }

		
	    int waitingThreads = 0;
	    for (int i = 0; i < threads.size(); i++){
		if (threads.get(i).getState() == Thread.State.valueOf("WAITING")) {
		    waitingThreads++;
		}
	    }
	    // If so, then we are done loading.
	    if (waitingThreads == threads.size()){
		// Pass the poison to kill the worker threads
		base.putPoison(poison);		
		done = true;
	    }	    
	}

	// Now wait for all the threads to return their results

	while (finishedThreadCount < maxThread) {
	    try {
		Global.monitor.writeToBuffer("BulkFingerprinter","computeAllFingerprints","awaiting thread results");
		Thread.sleep(100);
	    }
	    catch (InterruptedException e){
		ExceptionHandler.handleExceptionExit(e);
	    }
	}
		


	Global.monitor.writeToBuffer("BulkFingerprinter", "computeAllFingerprints", "complete");

	//System.err.println("BulkFingeprinter: return");
	return fpList;  
    
    }
    
    
    
}