package edu.georgetown.gucs.matcher;

import java.io.File;
import java.util.Map;
import java.util.SortedMap;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.List;
import java.util.Vector;
import java.util.Collections;
import java.util.ArrayList;
import java.util.Random;
import java.math.BigInteger;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.atomic.AtomicInteger;

import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.matcher.FingerprintMatcher;
import edu.georgetown.gucs.matcher.BulkMatcherWorker;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.ObjectBuilder;
import edu.georgetown.gucs.utility.Pair;
import edu.georgetown.gucs.utility.Token;

/**
 * This is a class that allows threaded parallel comparison of fingerprints.
 * 
 * @author Clay Shields
 **/

public class BulkMatcher {

    private String matcherName;
    private int minScore;
    private int maxThread;
    private List<Pair<byte[], Token>> fingerprints;
    // internal results is a vector from 0 to 100, one for each possible score
    // Each entry is a list of matching file pairs
    private List<List<Pair<Token,Token>>> internalResults;
    private SortedMap<Integer, List<String>> results;
    private LinkedBlockingQueue<Pair<Integer,Integer>> matchingQueue;
    private AtomicInteger finishedThreads;


    /**
     * A constructor for when you want to limit the number of threads that are run
     * @param matcherName
     *           the name of the fingerprinter-derived class to use for fingerprinting
     * @param results
     *      A map to insert the results in
     * @param minScore
     *      The minimum score required to be added to the results
     * @param maxThread
     *      limit on number of threads to be used
     */
    BulkMatcher(String matcherType, SortedMap<Integer, List<String>> results, int score, int maxThread){
	


	this.matcherName = matcherType;
	this.results = results;
	this.minScore = score;
	this.maxThread = maxThread;
	this.matchingQueue = new LinkedBlockingQueue<Pair<Integer,Integer>>();
	this.internalResults =  Collections.synchronizedList(new ArrayList<List<Pair<Token,Token>>> (100));

	for (int i = 0; i < 100; i++){
	    this.internalResults.add(i, Collections.synchronizedList(new ArrayList<Pair<Token,Token>>()));
	}

	finishedThreads = new AtomicInteger(0);
	
	Global.monitor.writeToBuffer("BulkMatcher","constructor","Matcher: " + this.matcherName + "min score: " + this.minScore + " maxThread: " + maxThread);

    }
    
    /**
     * A method to do n^2 comparisons among all fingeprints present
     *
     * @param fingerprints
     *           a list of the fingerprint/files pairs to compare     
     */
     

    public void compareAllFingerprints(List<Pair<byte[], Token>> fingerprints){
	
	this.fingerprints = fingerprints;
	
	Global.monitor.writeToBuffer("BulkMatcher","compareAllFingerprints","Number of fingerprints" +  fingerprints.size());

	Pair<Integer, Integer> poison = new Pair(new Integer(-1), new Integer(-1));
	ArrayList<BulkMatcherWorker> threads = new ArrayList<BulkMatcherWorker>(maxThread);	

	Global.monitor.writeToBuffer("BulkMatcher","compareAllFingerprints","Starting " + maxThread + " threads");

	for (int i = 0; i < maxThread; i++){
	    threads.add(i, new BulkMatcherWorker(this.matcherName,this.minScore,matchingQueue,this.fingerprints,this.internalResults,poison,finishedThreads));

	    threads.get(i).setName("Bulk Matcher Worker " + i);
	    threads.get(i).setDaemon(true);
	    threads.get(i).start();
	} 



	// Fill the queue with columns for each thread to run throuhg
	for (Integer i = 0; i < fingerprints.size() - 2 ; i ++) {
	    try {
		matchingQueue.put(new Pair(i,fingerprints.size()));
	    }
	    catch (InterruptedException e){
		ExceptionHandler.handleExceptionExit(e);
	    }
	}
    
	
	
	// Now wait for the threads to finish processing
	boolean done = false;
	while (!done){
	    // Sleep a bit while the threads run
	    try {
		Global.monitor.writeToBuffer("BulkMatcher","compareAllFingerprints","awaiting thread termination");
		Thread.sleep(100);
	    }
	    catch (InterruptedException e){
		ExceptionHandler.handleExceptionExit(e);
	    }

	    // If there are elements left, we are not done
	    if (matchingQueue.peek() == null){
		// If the elements are gone, wait to see how many threads are running
		int waitingThreads = 0;
		for (int i = 0; i < threads.size(); i++){
		    if (threads.get(i).getState() == Thread.State.valueOf("WAITING")) {
			waitingThreads++;
		    }
		}
		// If so, then we are done loading and matching.
		if (waitingThreads == threads.size()){
		    Global.monitor.writeToBuffer("BulkMatcher","compareAllFingerprints","terminating threads");
		    // Pass the poison to kill the worker threads
		    try {
			matchingQueue.put(poison);
		    }
		    catch (InterruptedException e){
			ExceptionHandler.handleExceptionExit(e);
		    }
		    
		    done = true;
		}	    
	    }

	}

	while (finishedThreads.get() < maxThread) {
	    try {
		Global.monitor.writeToBuffer("BulkMatcher","compareAllFingerprints","awaiting thread results");
		Thread.sleep(100);
	    }
	    catch (InterruptedException e){
		ExceptionHandler.handleExceptionExit(e);
	    }
	}
		
	    
	// Now we have to process the internal results into the format expected by the external results
	// 

	Global.monitor.writeToBuffer("BulkMatcher","compareAllFingerprints","Reformatting results");
	for (int i = 99; i > 0; i--){
	    List<Pair<Token,Token>> outerVec = internalResults.get(i);
	    if (outerVec != null){
		for (Pair<Token,Token> pair : outerVec){
		    //System.err.println (i + ": " + pair.getA().name() + " | " + pair.getB().name());
		
		    List<String> updatedScores;
		    if (this.results.containsKey(i)) {
			updatedScores = this.results.get(i);
		    } else {
			updatedScores = new Vector<String>();
		    }
		    updatedScores.add(String.format("%s%n%s", pair.getA().name() , pair.getB().name()));
		    this.results.put(i, updatedScores);

		}
	    }

	}

	return;
	
    }

}