package edu.georgetown.gucs.tokenizers;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Serializable;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Token;

/**
 * Splits a document into tokens (either by line or by word) or alters the tokens in various ways. This is the base
 * class for all tokenizers.
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 */
public class Tokenizer implements Serializable, Iterable<Entry<String, List<Token>>> {

  private static final long serialVersionUID = 2475577088718479996L;

  /** The map containing the splitter name and its corresponding list of token elements */
  protected Map<String, List<Token>> tokenVectorMap = Collections.synchronizedMap(new TreeMap<String, List<Token>>());

  public Tokenizer(){
      //System.err.println("Hi, I am a tokenizer for thread: " + Thread.currentThread().getName());
  }

  /**
   * Alters or eliminates certain tokens. By default, this method adds all string tokens from the given iterator to the
   * list of tokens, and must be over-ridden in each specific tokenizer.
   * 
   * @param tokenVector
   *          the vector of token elements to tokenize
   * 
   * @return the vector of remaining tokens after tokenization
   */
  public List<Token> tokenize(List<Token> tokenVector) {
    this.tokenVectorMap.put("null", tokenVector);
    return tokenVector;
  }

  /**
   * Splits the document into tokens. By default, this method calls the ExceptionHandler to throw a
   * "Tokenizer does not support reading from a file" exception. It must be over-ridden in each tokenizer that supports
   * reading from a file.
   * 
   * @param filename
   *          the filename of the document to split into tokens
   * 
   * @return the vector of tokens present in the file
   */
  public List<Token> tokenize(String filename) {
    ExceptionHandler.handleExceptionThrow(new Exception("Tokenizer does not support reading from a file"));
    return null;
  }

  /**
   * Provides the list of each token in order of its appearance
   * 
   * @return the map containing a splitter names and corresponding list of token elements
   */
  public Map<String, List<Token>> getTokenVectorMap() {
    return this.tokenVectorMap;
  }

  @Override
  public Iterator<Entry<String, List<Token>>> iterator() {
    synchronized (this.tokenVectorMap) {
      return this.tokenVectorMap.entrySet().iterator();
    }
  }

  /**
   * Prints each token to the system output stream using the UTF-8 charset
   */
  public void printTokens() {
    try {
      // TODO Tokenizer: Not sure this UTF-8 does anything
      OutputStreamWriter out = new OutputStreamWriter(System.out, "UTF-8");
      BufferedWriter writer = new BufferedWriter(out);
      for (Entry<String, List<Token>> entry : this.tokenVectorMap.entrySet()) {
        StringBuffer header = new StringBuffer("**********");
        for (int i = 0; i < entry.getKey().length(); i++) {
          header.append("*");
        }
        writer.write(String.format("%s%nSplitter: %s%n%s%n", header.toString(), entry.getKey(), header.toString()));
        for (Token token : entry.getValue()) {
          writer.write(String.format("%s%n", token));
        }
        writer.write(String.format("%n"));
      }
      out.flush();
      writer.flush();
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }
}
