package edu.georgetown.gucs.utility;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;

/**
 * Builds an Object of the given name
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 */
public class ObjectBuilder {

  /**
   * Builds an object with given name and parameters
   * 
   * @param objAndParams
   *          the string object name and parameters to use for construction
   * 
   * @return the created object
   */
  public static Object buildObject(String objAndParams) {
    if (objAndParams.contains("("))
      return buildObjectCommand(objAndParams);
    return buildObjectXML(objAndParams);
  }

  /**
   * Builds an object where object and parameters are separated by whitespace
   * 
   * @param objAndParams
   *          the string object name and parameters to use for construction
   * 
   * @return the created object
   */
  private static Object buildObjectXML(String objAndParams) {

    if (objAndParams == null) {
      ExceptionHandler.handleExceptionExit(new Exception("Object.buildObjectXML: null parameter"));
    } else {
      List<String> splitArray = Arrays.asList(objAndParams.split("\\s"));
      String classname = splitArray.get(0);
      Object[] parameterArray = new String[] {};
      Class<?>[] classType = new Class<?>[] {};
      
      try {
        if (splitArray.size() > 1) {
          parameterArray = splitArray.subList(1, splitArray.size()).toArray();
          classType = new Class[parameterArray.length];
          Arrays.fill(classType, Class.forName("java.lang.String"));
        }

        // Code from http://www.rgagnon.com/javadetails/java-0351.html
        Class<?> cl = Class.forName(getFullName(classname));
        Constructor<?> co = cl.getConstructor(classType);
        return co.newInstance(parameterArray);

      } catch (ClassNotFoundException e) {
	  
	  Global.monitor.writeToBuffer("ObjectBuilder","ClassNotFoundException",objAndParams);
        ExceptionHandler.handleExceptionExit(e);
      } catch (NoSuchMethodException e) {
        ExceptionHandler.handleExceptionExit(e);
      } catch (InstantiationException e) {
        ExceptionHandler.handleExceptionExit(e);
      } catch (IllegalAccessException e) {
        ExceptionHandler.handleExceptionExit(e);
      } catch (InvocationTargetException e) {
        ExceptionHandler.handleExceptionExit(e);
      }
    }
    return objAndParams;
  }

  /**
   * Builds an object that is declared as <code>Object(parameters...)</code>; separates object using parentheses and
   * parameters using commas
   * 
   * @param objAndParams
   *          the string object name and parameters to use for construction
   * 
   * @return the created object
   */
  private static Object buildObjectCommand(String objAndParams) {

    if (objAndParams == null) {
      ExceptionHandler.handleExceptionExit(new Exception("Object.buildObject: null parameter"));
    } else {
      String classname = objAndParams.substring(0, objAndParams.indexOf('(')).trim();
      Class<?>[] classType = new Class<?>[] {};
      Object[] parameterArray = new String[] {};
      
      try {
        if (objAndParams.indexOf('(') + 1 < objAndParams.indexOf(')')) {
          String parameters = objAndParams.substring(objAndParams.indexOf('(') + 1, objAndParams.indexOf(')'));
          parameterArray = parameters.split(",");
          classType = new Class[parameterArray.length];
          Arrays.fill(classType, Class.forName("java.lang.String"));
        }

        // Code from http://www.rgagnon.com/javadetails/java-0351.html
        Class<?> cl = Class.forName(getFullName(classname));
        Constructor<?> co = cl.getConstructor(classType);
        return co.newInstance(parameterArray);

      } catch (ClassNotFoundException e) {
        ExceptionHandler.handleExceptionExit(e);
      } catch (NoSuchMethodException e) {
        ExceptionHandler.handleExceptionExit(e);
      } catch (InstantiationException e) {
        ExceptionHandler.handleExceptionExit(e);
      } catch (IllegalAccessException e) {
        ExceptionHandler.handleExceptionExit(e);
      } catch (InvocationTargetException e) {
        ExceptionHandler.handleExceptionExit(e);
      }
    }
    return objAndParams;
  }

  /**
   * Provides the full package name for a given class
   * 
   * @param classname
   *          the string class name of the object to build
   * 
   * @return the string full package name of the class
   */
  private static String getFullName(String classname) {
    // This only works because of the naming convention used for classes
    if (classname.contains("edu.georgetown.gucs."))
      return classname;
    if (classname.toLowerCase(Locale.ENGLISH).contains("dictionary"))
      return "edu.georgetown.gucs.dictionary." + classname;
    if (classname.toLowerCase(Locale.ENGLISH).contains("fingerprinter"))
      return "edu.georgetown.gucs.fingerprinter." + classname;
    if (classname.toLowerCase(Locale.ENGLISH).contains("matcher"))
      return "edu.georgetown.gucs.matcher." + classname;
    if (classname.toLowerCase(Locale.ENGLISH).contains("tokenizer"))
      return "edu.georgetown.gucs.tokenizers." + classname;
    return classname;
  }
}
